import os
import pytest
import numpy as np
from os.path import join
from astropy.io.fits import Header
from .. import open as pyregion_open
from numpy.testing import assert_allclose

rootdir = join(os.path.dirname(os.path.abspath(__file__)), 'data')


@pytest.fixture(scope="module")
def header():
    return Header.fromtextfile(join(rootdir, "sample_fits01.header"))


@pytest.mark.parametrize(("ref_name", "reg_name", "header_name"), [
    ("test01_img.reg", "test01_fk5_sexagecimal.reg", "sample_fits01.header"),
    ("test01_img.reg", "test01_gal.reg", "sample_fits01.header"),
    ("test01_img.reg", "test01_ds9_physical.reg", "sample_fits01.header"),
    ("test01_img.reg", "test01_fk5_degree.reg", "sample_fits01.header"),
    ("test01_img.reg", "test01_mixed.reg", "sample_fits01.header"),
    ("test01_img.reg", "test01_ciao.reg", "sample_fits01.header"),
    ("test01_img.reg", "test01_ciao_physical.reg", "sample_fits01.header"),
    ("test01_img.reg", "test01_fk5.reg", "sample_fits01.header"),
    ("test01_img.reg", "test01_fk4.reg", "sample_fits01.header"),
    ("test01_img.reg", "test01_icrs.reg", "sample_fits01.header"),
    ("test02_1_img.reg", "test02_1_fk5.reg", "sample_fits02.header"),
    ("test_annuli.reg", "test_annuli_wcs.reg", "sample_fits01.header"),
    ("test03_img.reg", "test03_fk5.reg", "sample_fits03.header"),
    ("test03_img.reg", "test03_icrs.reg", "sample_fits03.header"),
    ("test03_img.reg", "test03_ciao_physical.reg", "sample_fits03.header"),
    ("test03_img.reg", "test03_gal.reg", "sample_fits03.header"),
])
def test_region(ref_name, reg_name, header_name):
    header = Header.fromtextfile(join(rootdir, header_name))
    ref_region = pyregion_open(join(rootdir, ref_name)).as_imagecoord(header)

    r = pyregion_open(join(rootdir, reg_name)).as_imagecoord(header)

    assert len(r) == len(ref_region)

    for ref_reg, reg in zip(ref_region, r):
        if reg.name == "rotbox":
            reg.name = "box"

        assert ref_reg.name == reg.name

        # Normalize everything like angles
        ref_list = np.asarray(ref_reg.coord_list)
        reg_list = np.asarray(reg.coord_list)
        assert_allclose((ref_list + 180) % 360 - 180,
                        (reg_list + 180) % 360 - 180,
                        atol=0.03)

        assert ref_reg.exclude == reg.exclude


@pytest.mark.parametrize("reg_name", [
    "test_annuli_ciao.reg",  # subset of test03_img.reg
    "test_context.reg",
    "test02.reg",
    "test04_img.reg",
    "test_text.reg",
    "test01.reg",
])
def test_open_regions(reg_name, header):
    # TODO: Better test. Like figure out how these files relate to each other
    pyregion_open(join(rootdir, reg_name)).as_imagecoord(header)



import copy
import numpy as np
from math import cos, sin, pi, atan2
import warnings
import matplotlib.patches as mpatches
from matplotlib.path import Path
from matplotlib.lines import Line2D
from matplotlib.transforms import Affine2D, Bbox, IdentityTransform
from matplotlib.text import Annotation


def rotated_polygon(xy, ox, oy, angle):
    # angle in degree
    theta = angle / 180. * pi

    st = sin(theta)
    ct = cos(theta)

    xy = np.asarray(xy, dtype="d")
    x, y = xy[:, 0], xy[:, 1]
    x1 = x - ox
    y1 = y - oy

    x2 = ct * x1 + -st * y1
    y2 = st * x1 + ct * y1

    xp = x2 + ox
    yp = y2 + oy

    return np.hstack((xp.reshape((-1, 1)), yp.reshape((-1, 1))))

    # sss3 = [s1[0] for s1 in sss2 if isinstance(s1[0], parser_ds9.Shape)]


_point_type_dict = dict(circle="o",
                        box="s",
                        diamond="D",
                        x="x",
                        cross="+",
                        arrow="^",
                        boxcircle="*")

_ds9_to_mpl_colormap = dict(green="lime",
                            )


def properties_func_default(shape, saved_attrs):
    attr_list = copy.copy(shape.attr[0])
    attr_dict = copy.copy(shape.attr[1])

    attr_list.extend(saved_attrs[0])
    attr_dict.update(saved_attrs[1])

    color = attr_dict.get("color", None)
    color = _ds9_to_mpl_colormap.get(color, color)

    if shape.name == "text":
        kwargs = dict(color=color,
                      rotation=attr_dict.get("textangle", 0),
                      )
        font = attr_dict.get("font")
        if font:
            a = font.split()
            if len(a) >= 3:
                fontsize = float(a[1])
                kwargs["fontsize"] = fontsize
    elif shape.name == "point":
        point_attrs = attr_dict.get("point", "boxcircle").split()
        if len(point_attrs) == 1:
            point_type = point_attrs[0]
            point_size = 11
        elif len(point_attrs) > 1:
            point_type = point_attrs[0]
            point_size = int(point_attrs[1])

        marker = _point_type_dict.get(point_type, "o")
        kwargs = dict(markeredgecolor=color,
                      markerfacecolor="none",
                      marker=marker,
                      markeredgewidth=int(attr_dict.get("width", 1)),
                      markersize=point_size
                      )
    elif shape.name in ["line", "vector"]:
        fontsize = 10  # default font size

        font = attr_dict.get("font")
        if font:
            a = font.split()
            if len(a) >= 3:
                fontsize = float(a[1])

        kwargs = dict(color=color,
                      linewidth=int(attr_dict.get("width", 1)),
                      mutation_scale=fontsize,
                      )
        if int(attr_dict.get("dash", "0")):
            kwargs["linestyle"] = "dashed"

    else:
        kwargs = dict(edgecolor=color,
                      linewidth=int(attr_dict.get("width", 1)),
                      facecolor="none"
                      )

        if "background" in attr_list:
            kwargs["linestyle"] = "dashed"

        if int(attr_dict.get("dash", "0")):
            kwargs["linestyle"] = "dashed"
        if shape.exclude:
            kwargs["hatch"] = "/"

    return kwargs


def _get_text(txt, x, y, dx, dy, ha="center", va="center", **kwargs):
    if "color" in kwargs:
        textcolor = kwargs["color"]
        del kwargs["color"]
    elif "markeredgecolor" in kwargs:
        textcolor = kwargs["markeredgecolor"]
    else:
        import matplotlib as mpl
        textcolor = mpl.rcParams['text.color']
    ann = Annotation(txt, (x, y), xytext=(dx, dy),
                     xycoords='data',
                     textcoords="offset points",
                     color=textcolor,
                     ha=ha, va=va,
                     **kwargs)
    ann.set_transform(IdentityTransform())

    return ann


def as_mpl_artists(shape_list,
                   properties_func=None,
                   text_offset=5.0, origin=1):
    """
    Converts a region list to a list of patches and a list of artists.


    Optional Keywords:
    [ text_offset ] - If there is text associated with the regions, add
    some vertical offset (in pixels) to the text so that it doesn't overlap
    with the regions.

    Often, the regions files implicitly assume the lower-left corner
    of the image as a coordinate (1,1). However, the python convetion
    is that the array index starts from 0. By default (origin = 1),
    coordinates of the returned mpl artists have coordinate shifted by
    (1, 1). If you do not want this shift, set origin=0.
    """

    patch_list = []
    artist_list = []

    if properties_func is None:
        properties_func = properties_func_default

    # properties for continued(? multiline?) regions
    saved_attrs = None

    for shape in shape_list:

        patches = []

        if saved_attrs is None:
            _attrs = [], {}
        else:
            _attrs = copy.copy(saved_attrs[0]), copy.copy(saved_attrs[1])

        kwargs = properties_func(shape, _attrs)

        if shape.name == "composite":
            saved_attrs = shape.attr
            continue

        if saved_attrs is None and shape.continued:
            saved_attrs = shape.attr
        #         elif (shape.name in shape.attr[1]):
        #             if (shape.attr[1][shape.name] != "ignore"):
        #                 saved_attrs = shape.attr

        if not shape.continued:
            saved_attrs = None

        # text associated with the shape
        txt = shape.attr[1].get("text")

        if shape.name == "polygon":
            xy = np.array(shape.coord_list)
            xy.shape = -1, 2

            # -1 for change origin to 0,0
            patches = [mpatches.Polygon(xy - origin, closed=True, **kwargs)]

        elif shape.name == "rotbox" or shape.name == "box":
            xc, yc, w, h, rot = shape.coord_list
            # -1 for change origin to 0,0
            xc, yc = xc - origin, yc - origin
            _box = np.array([[-w / 2., -h / 2.],
                             [-w / 2., h / 2.],
                             [w / 2., h / 2.],
                             [w / 2., -h / 2.]])
            box = _box + [xc, yc]
            rotbox = rotated_polygon(box, xc, yc, rot)
            patches = [mpatches.Polygon(rotbox, closed=True, **kwargs)]

        elif shape.name == "ellipse":
            xc, yc = shape.coord_list[:2]
            # -1 for change origin to 0,0
            xc, yc = xc - origin, yc - origin
            angle = shape.coord_list[-1]

            maj_list, min_list = shape.coord_list[2:-1:2], shape.coord_list[3:-1:2]

            patches = [mpatches.Ellipse((xc, yc), 2 * maj, 2 * min,
                                        angle=angle, **kwargs)
                       for maj, min in zip(maj_list, min_list)]

        elif shape.name == "annulus":
            xc, yc = shape.coord_list[:2]
            # -1 for change origin to 0,0
            xc, yc = xc - origin, yc - origin
            r_list = shape.coord_list[2:]

            patches = [mpatches.Ellipse((xc, yc), 2 * r, 2 * r, **kwargs) for r in r_list]

        elif shape.name == "circle":
            xc, yc, major = shape.coord_list
            # -1 for change origin to 0,0
            xc, yc = xc - origin, yc - origin
            patches = [mpatches.Ellipse((xc, yc), 2 * major, 2 * major, angle=0, **kwargs)]

        elif shape.name == "panda":
            xc, yc, a1, a2, an, r1, r2, rn = shape.coord_list
            # -1 for change origin to 0,0
            xc, yc = xc - origin, yc - origin
            patches = [mpatches.Arc((xc, yc), rr * 2, rr * 2, angle=0,
                                    theta1=a1, theta2=a2, **kwargs)
                       for rr in np.linspace(r1, r2, rn + 1)]

            for aa in np.linspace(a1, a2, an + 1):
                xx = np.array([r1, r2]) * np.cos(aa / 180. * np.pi) + xc
                yy = np.array([r1, r2]) * np.sin(aa / 180. * np.pi) + yc
                p = Path(np.transpose([xx, yy]))
                patches.append(mpatches.PathPatch(p, **kwargs))

        elif shape.name == "pie":
            xc, yc, r1, r2, a1, a2 = shape.coord_list
            # -1 for change origin to 0,0
            xc, yc = xc - origin, yc - origin

            patches = [mpatches.Arc((xc, yc), rr * 2, rr * 2, angle=0,
                                    theta1=a1, theta2=a2, **kwargs)
                       for rr in [r1, r2]]

            for aa in [a1, a2]:
                xx = np.array([r1, r2]) * np.cos(aa / 180. * np.pi) + xc
                yy = np.array([r1, r2]) * np.sin(aa / 180. * np.pi) + yc
                p = Path(np.transpose([xx, yy]))
                patches.append(mpatches.PathPatch(p, **kwargs))

        elif shape.name == "epanda":
            xc, yc, a1, a2, an, r11, r12, r21, r22, rn, angle = shape.coord_list
            # -1 for change origin to 0,0
            xc, yc = xc - origin, yc - origin

            # mpl takes angle a1, a2 as angle as in circle before
            # transformation to ellipse.

            x1, y1 = cos(a1 / 180. * pi), sin(a1 / 180. * pi) * r11 / r12
            x2, y2 = cos(a2 / 180. * pi), sin(a2 / 180. * pi) * r11 / r12

            a1, a2 = atan2(y1, x1) / pi * 180., atan2(y2, x2) / pi * 180.

            patches = [mpatches.Arc((xc, yc), rr1 * 2, rr2 * 2,
                                    angle=angle, theta1=a1, theta2=a2,
                                    **kwargs)
                       for rr1, rr2 in zip(np.linspace(r11, r21, rn + 1),
                                           np.linspace(r12, r22, rn + 1))]

            for aa in np.linspace(a1, a2, an + 1):
                xx = np.array([r11, r21]) * np.cos(aa / 180. * np.pi)
                yy = np.array([r11, r21]) * np.sin(aa / 180. * np.pi)
                p = Path(np.transpose([xx, yy]))
                tr = Affine2D().scale(1, r12 / r11).rotate_deg(angle).translate(xc, yc)
                p2 = tr.transform_path(p)
                patches.append(mpatches.PathPatch(p2, **kwargs))

        elif shape.name == "text":
            xc, yc = shape.coord_list[:2]
            # -1 for change origin to 0,0
            xc, yc = xc - origin, yc - origin

            if txt:
                _t = _get_text(txt, xc, yc, 0, 0, **kwargs)
                artist_list.append(_t)

        elif shape.name == "point":
            xc, yc = shape.coord_list[:2]
            # -1 for change origin to 0,0
            xc, yc = xc - origin, yc - origin
            artist_list.append(Line2D([xc], [yc],
                                      **kwargs))

            if txt:
                textshape = copy.copy(shape)
                textshape.name = "text"
                textkwargs = properties_func(textshape, _attrs)
                _t = _get_text(txt, xc, yc, 0, text_offset,
                               va="bottom",
                               **textkwargs)
                artist_list.append(_t)

        elif shape.name in ["line", "vector"]:
            if shape.name == "line":
                x1, y1, x2, y2 = shape.coord_list[:4]
                # -1 for change origin to 0,0
                x1, y1, x2, y2 = x1 - origin, y1 - origin, x2 - origin, y2 - origin

                a1, a2 = shape.attr[1].get("line", "0 0").strip().split()[:2]

                arrowstyle = "-"
                if int(a1):
                    arrowstyle = "<" + arrowstyle
                if int(a2):
                    arrowstyle = arrowstyle + ">"

            else:  # shape.name == "vector"
                x1, y1, l, a = shape.coord_list[:4]
                # -1 for change origin to 0,0
                x1, y1 = x1 - origin, y1 - origin
                x2, y2 = x1 + l * np.cos(a / 180. * np.pi), y1 + l * np.sin(a / 180. * np.pi)
                v1 = int(shape.attr[1].get("vector", "0").strip())

                if v1:
                    arrowstyle = "->"
                else:
                    arrowstyle = "-"

            patches = [mpatches.FancyArrowPatch(posA=(x1, y1),
                                                posB=(x2, y2),
                                                arrowstyle=arrowstyle,
                                                arrow_transmuter=None,
                                                connectionstyle="arc3",
                                                patchA=None, patchB=None,
                                                shrinkA=0, shrinkB=0,
                                                connector=None,
                                                **kwargs)]

        else:
            warnings.warn("'as_mpl_artists' does not know how to convert {0} "
                          "to mpl artist".format(shape.name))

        patch_list.extend(patches)

        if txt and patches:
            # the text associated with a shape uses different
            # matplotlib keywords than the shape itself for, e.g.,
            # color
            textshape = copy.copy(shape)
            textshape.name = "text"
            textkwargs = properties_func(textshape, _attrs)

            # calculate the text position
            _bb = [p.get_window_extent() for p in patches]

            # this is to work around backward-incompatible change made
            # in matplotlib 1.2. This change is later reverted so only
            # some versions are affected. With affected version of
            # matplotlib, get_window_extent method calls get_transform
            # method which sets the _transformSet to True, which is
            # not desired.
            for p in patches:
                p._transformSet = False

            _bbox = Bbox.union(_bb)
            x0, y0, x1, y1 = _bbox.extents
            xc = .5 * (x0 + x1)

            _t = _get_text(txt, xc, y1, 0, text_offset,
                           va="bottom",
                           **textkwargs)
            artist_list.append(_t)

    return patch_list, artist_list



from ..region_numbers import (usn, simple_integer, sexadecimal60, Sixty,
                              hms_number, dms_number, angular_distance,
                              CoordOdd, HMS)


def test_usn():
    for f in ["32.4", "0.23", "0.3e-7", "1.234e+7"]:
        assert usn.parseString(f)[0] == f


def test_integer():
    for f in ["32", "+3"]:
        assert simple_integer.parseString(f)[0].text == f


def test_sexadecimal():
    s = sexadecimal60.parseString

    assert s("32:24:32.2")[0].v == Sixty(1, 32, 24, 32.2).v
    assert s("-32:24:32.2")[0].v == Sixty(-1, 32, 24, 32.2).v
    assert s("+32:24:32.2")[0].v == Sixty(1, 32, 24, 32.2).v


def test_hms():
    s = hms_number.parseString

    assert s("32h24m32.2s")[0].v == Sixty(1, 32, 24, 32.2).v
    assert s("0h24m32.2s")[0].v == Sixty(1, 0, 24, 32.2).v
    assert s("32h")[0].v == Sixty(1, 32, 0, 0).v


def test_dms():
    s = dms_number.parseString

    assert s("32d24m32.2s")[0].v == Sixty(1, 32, 24, 32.2).v
    assert s("-32d24m32.2s")[0].v == Sixty(-1, 32, 24, 32.2).v
    assert s("32d")[0].v == Sixty(1, 32, 0, 0).v


def test_ang_distance():
    s = angular_distance.parseString

    assert s("32.3'")[0].v == Sixty(1, 0, 32.3, 0.).v
    assert s("32\'24\"")[0].v == Sixty(1, 0, 32, 24).v


def test_coord_odd():
    s = CoordOdd.parser.parseString

    assert s("32h24m32.2s")[0].v == Sixty(1, 32, 24, 32.2).v
    assert s("32:24:32.2s")[0].v == Sixty(1, 32, 24, 32.2).v
    assert s("32.24")[0].v == 32.24

    s1 = s("32:24:32.2s")[0]
    assert isinstance(s1, HMS)



from pyparsing import Literal, Optional, Combine, Or, Word, nums


def _unsigned_simple_number():
    # fnumber : 102.43, 12304.3e10,
    #           .32???
    point = Literal(".")
    e = Literal("e") | Literal("E")  # CaselessLiteral( "E" )
    fnumber = Combine(Word(nums) +
                      Optional(point + Optional(Word(nums))) +
                      Optional(e + Word("+-" + nums, nums)))

    return fnumber  # .setParseAction(lambda s,l,t: (float(t[0]), t[0]))


usn = _unsigned_simple_number()


def get_default(v):
    def _f(s, l, tok):
        if tok:
            return tok
        return v

    return _f


optional_sign = Optional(Literal("+") | Literal("-")).setParseAction(get_default(""))


class SimpleNumber(object):
    def __repr__(self):
        return "Number(%s)" % (self.text,)

    def __str__(self):
        return self.__repr__()

    def __init__(self, text):
        self.text = text
        self.v = float(text)


def _simple_number():
    s = Combine(optional_sign + usn)
    s.setParseAction(lambda s, l, tok: SimpleNumber(tok[0]))

    return s


simple_number = _simple_number()


class SimpleInteger(object):
    def __repr__(self):
        return "Number(%s)" % (self.text,)

    def __str__(self):
        return self.__repr__()

    def __init__(self, text):
        self.text = text
        self.v = int(text)


def _unsigned_integer():
    s = Combine(Optional("+") + Word(nums))
    s.setParseAction(lambda s, l, tok: SimpleInteger(tok[0]))

    return s


simple_integer = _unsigned_integer()


class Sixty(object):
    def __init__(self, sn, d, m, s):
        self.v = sn * (d + (m + s / 60.) / 60.)
        self.degree = self.v


class HMS(object):
    def __repr__(self):
        return "HMS(%s)" % (self.text,)

    def __init__(self, kl):
        self.text = "".join(kl)

        if kl[0] == "-":
            sn = -1
        else:
            sn = +1

        kkl = kl[1::2]
        if len(kkl) == 3:
            d, m, s = float(kl[1]), float(kl[3]), float(kl[5])
        elif len(kkl) == 2:
            d, m, s = float(kl[1]), float(kl[3]), 0.
        else:
            d, m, s = float(kl[1]), 0., 0.

        self.v = sn * (d + (m + s / 60.) / 60.)
        self.degree = self.v * 15


class DMS(object):
    def __repr__(self):
        return "DMS(%s)" % (self.text,)

    def __init__(self, kl):
        self.text = "".join(kl)

        if kl[0] == "-":
            sn = -1
        else:
            sn = +1

        kkl = kl[1::2]
        if len(kkl) == 3:
            d, m, s = float(kl[1]), float(kl[3]), float(kl[5])
        elif len(kkl) == 2:
            d, m, s = float(kl[1]), float(kl[3]), 0.
        else:
            d, m, s = float(kl[1]), 0., 0.

        self.v = sn * (d + (m + s / 60.) / 60.)
        self.degree = self.v


class AngularDistance(object):
    def __repr__(self):
        return "Ang(%s)" % (self.text,)

    def __init__(self, kl):
        self.text = "".join(kl)

        m, s = 0, 0
        if kl[1] == "'":
            m = float(kl[0])
            if len(kl) == 4:
                s = float(kl[2])
        else:
            s = float(kl[0])

        self.v = (m + s / 60.) / 60.
        self.degree = self.v


def _sexadecimal():
    colon = Literal(":")

    s = optional_sign + usn + colon + usn + \
        Optional(colon + usn)

    return s


sexadecimal60 = _sexadecimal().setParseAction(lambda s, l, tok: DMS(tok))
sexadecimal24 = _sexadecimal().setParseAction(lambda s, l, tok: HMS(tok))


def _hms_number():
    _h = (usn + Literal("h")).leaveWhitespace()
    _m = (usn + Literal("m")).leaveWhitespace()
    _s = (usn + Literal("s")).leaveWhitespace()

    hms = optional_sign + _h + Optional(_m + Optional(_s))

    hms = hms.setParseAction(lambda s, l, tok: HMS(tok))

    return hms


def _dms_number():
    _d = (usn + Literal("d")).leaveWhitespace()
    _m = (usn + Literal("m")).leaveWhitespace()
    _s = (usn + Literal("s")).leaveWhitespace()

    dms = optional_sign + _d + Optional(_m + Optional(_s))

    dms = dms.setParseAction(lambda s, l, tok: DMS(tok))

    return dms


hms_number = _hms_number()
dms_number = _dms_number()


def _angular_distance():
    _m = (usn + Literal("\'").leaveWhitespace())
    _s = (usn + Literal("\"").leaveWhitespace())

    ms = Or([_m + Optional(_s), _s])

    ms = ms.setParseAction(lambda s, l, tok: AngularDistance(tok))

    return ms


angular_distance = _angular_distance()


class Arg(object):
    def __init__(self, type, parser):
        self.type = type
        self.parser = parser


class CoordOdd:
    parser = (hms_number | sexadecimal24 | simple_number)
    type = HMS


class CoordEven:
    parser = (dms_number | sexadecimal60 | simple_number)
    type = DMS


class Distance:
    parser = (angular_distance | simple_number)
    type = AngularDistance


class Angle:
    parser = (simple_number)
    type = simple_number


class Integer:
    parser = simple_integer
    type = simple_number



import copy

from astropy.coordinates import SkyCoord
from astropy.wcs import WCS
from astropy.wcs.utils import proj_plane_pixel_area, proj_plane_pixel_scales
import numpy as np
from .wcs_helper import _estimate_angle
from .region_numbers import CoordOdd, Distance, Angle
from .parser_helper import Shape, CoordCommand
from .region_numbers import SimpleNumber, SimpleInteger


def _generate_arg_types(coordlist_length, shape_name):
    """Find coordinate types based on shape name and coordlist length

    This function returns a list of coordinate types based on which
    coordinates can be repeated for a given type of shap

    Parameters
    ----------
    coordlist_length : int
        The number of coordinates or arguments used to define the shape.

    shape_name : str
        One of the names in `pyregion.ds9_shape_defs`.

    Returns
    -------
    arg_types : list
        A list of objects from `pyregion.region_numbers` with a length equal to
        coordlist_length.

    """
    from .ds9_region_parser import ds9_shape_defs
    from .ds9_attr_parser import ds9_shape_in_comment_defs

    if shape_name in ds9_shape_defs:
        shape_def = ds9_shape_defs[shape_name]
    else:
        shape_def = ds9_shape_in_comment_defs[shape_name]

    initial_arg_types = shape_def.args_list
    arg_repeats = shape_def.args_repeat

    if arg_repeats is None:
        return initial_arg_types

    # repeat args between n1 and n2
    n1, n2 = arg_repeats
    arg_types = list(initial_arg_types[:n1])
    num_of_repeats = coordlist_length - (len(initial_arg_types) - n2)
    arg_types.extend((num_of_repeats - n1) //
                     (n2 - n1) * initial_arg_types[n1:n2])
    arg_types.extend(initial_arg_types[n2:])
    return arg_types


def convert_to_imagecoord(shape, header):
    """Convert the coordlist of `shape` to image coordinates

    Parameters
    ----------
    shape : `pyregion.parser_helper.Shape`
        The `Shape` to convert coordinates

    header : `~astropy.io.fits.Header`
        Specifies what WCS transformations to use.

    Returns
    -------
    new_coordlist : list
        A list of image coordinates defining the shape.

    """
    arg_types = _generate_arg_types(len(shape.coord_list), shape.name)

    new_coordlist = []
    is_even_distance = True
    coord_list_iter = iter(zip(shape.coord_list, arg_types))

    new_wcs = WCS(header)
    pixel_scales = proj_plane_pixel_scales(new_wcs)

    for coordinate, coordinate_type in coord_list_iter:
        if coordinate_type == CoordOdd:
            even_coordinate = next(coord_list_iter)[0]

            old_coordinate = SkyCoord(coordinate, even_coordinate,
                                      frame=shape.coord_format, unit='degree',
                                      obstime='J2000')
            new_coordlist.extend(
                np.asscalar(x)
                for x in old_coordinate.to_pixel(new_wcs, origin=1)
            )

        elif coordinate_type == Distance:
            if arg_types[-1] == Angle:
                degree_per_pixel = pixel_scales[0 if is_even_distance else 1]

                is_even_distance = not is_even_distance
            else:
                degree_per_pixel = np.sqrt(proj_plane_pixel_area(new_wcs))

            new_coordlist.append(coordinate / degree_per_pixel)

        elif coordinate_type == Angle:
            new_angle = _estimate_angle(coordinate,
                                        shape.coord_format,
                                        header)
            new_coordlist.append(new_angle)

        else:
            new_coordlist.append(coordinate)

    return new_coordlist


def convert_physical_to_imagecoord(shape, header):
    arg_types = _generate_arg_types(len(shape.coord_list), shape.name)

    new_coordlist = []
    coord_list_iter = iter(zip(shape.coord_list, arg_types))

    from .physical_coordinate import PhysicalCoordinate
    pc = PhysicalCoordinate(header)

    for coordinate, coordinate_type in coord_list_iter:
        if coordinate_type == CoordOdd:
            even_coordinate = next(coord_list_iter)[0]

            xy0 = pc.to_image(coordinate, even_coordinate)
            new_coordlist.extend(xy0)
        elif coordinate_type == Distance:
            new_coordlist.append(pc.to_image_distance(coordinate))
        else:
            new_coordlist.append(coordinate)

    return new_coordlist


def check_wcs_and_convert(args, all_dms=False):
    is_wcs = False

    value_list = []
    for a in args:
        if isinstance(a, SimpleNumber) or isinstance(a, SimpleInteger) \
                or all_dms:
            value_list.append(a.v)
        else:
            value_list.append(a.degree)
            is_wcs = True

    return is_wcs, value_list


def check_wcs(l):
    default_coord = "physical"

    for l1, c1 in l:
        if isinstance(l1, CoordCommand):
            default_coord = l1.text.lower()
            continue
        if isinstance(l1, Shape):
            if default_coord == "galactic":
                is_wcs, coord_list = check_wcs_and_convert(l1.params,
                                                           all_dms=True)
            else:
                is_wcs, coord_list = check_wcs_and_convert(l1.params)

            if is_wcs and (default_coord == "physical"):  # ciao format
                coord_format = "fk5"
            else:
                coord_format = default_coord

            l1n = copy.copy(l1)

            l1n.coord_list = coord_list
            l1n.coord_format = coord_format

            yield l1n, c1
        else:
            yield l1, c1



class PhysicalCoordinate(object):
    def __init__(self, header):
        phys_coord = ""

        # check if physical coordinate is defined. FIXME!
        for C in ["P", "L"]:
            try:
                if (header["WCSTY1" + C].strip() == "PHYSICAL") \
                        and (header["WCSTY2" + C].strip() == "PHYSICAL"):
                    phys_coord = C
            except KeyError:
                pass

            try:
                if (header["CTYPE1" + C].strip() == "X") \
                        and (header["CTYPE2" + C].strip() == "Y"):
                    phys_coord = C
            except KeyError:
                pass

        if phys_coord:
            C = phys_coord
            cv1, cr1, cd1 = header["CRVAL1" + C], header["CRPIX1" + C], header[" CDELT1" + C]
            cv2, cr2, cd2 = header["CRVAL2" + C], header["CRPIX2" + C], header[" CDELT2" + C]

            self._physical_coord_not_defined = False

            self.cv1_cr1_cd1 = cv1, cr1, cd1
            self.cv2_cr2_cd2 = cv2, cr2, cd2
            self.cdelt = (cd1 * cd2) ** .5

        else:
            self._physical_coord_not_defined = True
            self.cv1_cr1_cd1 = 0, 0, 1
            self.cv2_cr2_cd2 = 0, 0, 1
            self.cdelt = 1

    def to_physical(self, imx, imy):

        if self._physical_coord_not_defined:
            return imx, imy

        cv1, cr1, cd1 = self.cv1_cr1_cd1
        cv2, cr2, cd2 = self.cv2_cr2_cd2

        phyx = cv1 + (imx - cr1) * cd1
        phyy = cv2 + (imy - cr2) * cd2

        return phyx, phyy

    def to_image(self, phyx, phyy):

        if self._physical_coord_not_defined:
            return phyx, phyy

        cv1, cr1, cd1 = self.cv1_cr1_cd1
        cv2, cr2, cd2 = self.cv2_cr2_cd2

        imx = cr1 + (phyx - cv1) / cd1
        imy = cr2 + (phyy - cv2) / cd2

        return imx, imy

    def to_physical_distance(self, im_distance):

        if self._physical_coord_not_defined:
            return im_distance

        return im_distance * self.cdelt

    def to_image_distance(self, im_physical):

        if self._physical_coord_not_defined:
            return im_physical

        return im_physical / self.cdelt



import pytest
import os.path
from numpy.testing import assert_allclose
from astropy.io.fits import Header
from ..ds9_region_parser import ds9_shape_defs
from ..region_numbers import CoordOdd, CoordEven
from .. import wcs_converter
from ..wcs_helper import _calculate_rotation_angle

rootdir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')


def test__generate_arg_types_min_list():
    for name, shape in ds9_shape_defs.items():
        args_list = shape.args_list
        min_list = wcs_converter._generate_arg_types(len(args_list), name)
        for expected_arg, tested_arg in zip(args_list, min_list):
            assert expected_arg == tested_arg


@pytest.mark.parametrize(("name", "length", "result"), [
    ("polygon", 6, 3 * [CoordOdd, CoordEven]),
])
def test__generate_arg_types_with_repeats(name, length, result):
    test_list = wcs_converter._generate_arg_types(length, name)

    for expected_arg, tested_arg in zip(result, test_list):
        assert expected_arg == tested_arg


@pytest.mark.parametrize(("region_frame", "header_name", "rot_angle"), [
    ('fk5', 'sample_fits01.header', 0.00505712),
    ('galactic', 'sample_fits01.header', -19.2328),
    ('fk4', 'sample_fits01.header', -0.0810223),
    ('fk5', 'sample_fits03.header', -2.16043),
])
def test_calculate_rotation_angle(region_frame, header_name, rot_angle):
    header = Header.fromtextfile(os.path.join(rootdir, header_name))
    assert_allclose(
        _calculate_rotation_angle(region_frame, header), rot_angle,
        atol=0.001
    )



import os
import numpy as np
from os.path import join
from astropy.io.fits import Header
from .. import open as pyregion_open

rootdir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')


def demo_header():
    return Header.fromtextfile(join(rootdir, "sample_fits01.header"))


def test_region():
    ref_name = "test01_img.reg"
    header = demo_header()

    ref_region = pyregion_open(join(rootdir, ref_name)).as_imagecoord(header)
    mask = ref_region.get_mask(shape=(100, 100))

    assert isinstance(mask, np.ndarray) and mask.shape == (100, 100)

    # TODO: assert the content of the mask, too



from pyparsing import (Literal, CaselessKeyword, Optional, OneOrMore,
                       ZeroOrMore, restOfLine, MatchFirst, And, Or)


def as_comma_separated_list(al):
    l = [al[0]]
    comma = Literal(",").suppress()

    for a1 in al[1:]:
        l.append(comma)
        l.append(a1)

    return And(l)


class wcs_shape(object):
    def __init__(self, *kl, **kw):
        self.args_list = kl
        self.args_repeat = kw.get("repeat", None)

    def get_pyparsing(self):
        return [a.parser for a in self.args_list]


def define_shape(name, shape_args, args_repeat=None):
    lparen = Literal("(").suppress()
    rparen = Literal(")").suppress()
    comma = Literal(",").suppress()

    shape_name = CaselessKeyword(name)

    if args_repeat is None:
        shape_with_parens = And([shape_name, lparen,
                                 as_comma_separated_list(shape_args),
                                 rparen])

        shape_with_spaces = shape_name + And(shape_args)

    else:
        n1, n2 = args_repeat
        sl = []

        ss = shape_args[:n1]
        if ss:
            sl.append(as_comma_separated_list(ss))

        ss = shape_args[n1:n2]
        if ss:
            ar = as_comma_separated_list(ss)
            if sl:
                sl.extend([comma + ar, ZeroOrMore(comma + ar)])
            else:
                sl.extend([ar, ZeroOrMore(comma + ar)])

        ss = shape_args[n2:]
        if ss:
            if sl:
                sl.extend([comma, as_comma_separated_list(ss)])
            else:
                sl.extend([as_comma_separated_list(ss)])

        sl = [shape_name, lparen] + sl + [rparen]

        shape_with_parens = And(sl)

        shape_with_spaces = shape_name + OneOrMore(And(shape_args))

    return (shape_with_parens | shape_with_spaces)


def define_shape_helper(shape_defs):
    l = []

    for n, args in shape_defs.items():
        s = define_shape(n,
                         args.get_pyparsing(),
                         args_repeat=args.args_repeat)
        l.append(s)

    return Or(l)


def define_expr(regionShape, negate_func):
    minus = Literal("-").suppress()
    regionExclude = (minus + regionShape).setParseAction(negate_func)
    regionExpr = (regionShape | regionExclude)

    return regionExpr


def define_line(atom,
                separator,
                comment):
    atomSeparator = separator.suppress()

    atomList = atom + ZeroOrMore(atomSeparator + atom) + \
               Optional(atomSeparator)

    line = (atomList + Optional(comment)) | comment

    return line


def comment_shell_like(comment_begin, parseAction=None):
    c = comment_begin + restOfLine
    if parseAction:
        c = c.setParseAction(parseAction)

    return c


def define_simple_literals(literal_list, parseAction=None):
    l = MatchFirst([CaselessKeyword(k) for k in literal_list])

    if parseAction:
        l = l.setParseAction(parseAction)

    return l


class Shape(object):
    """Shape.

    Parameters
    ----------
    shape_name : str
        Shape name
    shape_params : list
        List of parameters

    Examples
    --------
    >>> import pyregion
    >>> region_string = 'fk5;circle(290.96388,14.019167,843.31194")'
    >>> shape_list = pyregion.parse(region_string)
    >>> shape = shape_list[0]
    >>> print(shape.__dict__)
    {'attr': ([], {}),
     'comment': None,
     'continued': None,
     'coord_format': 'fk5',
     'coord_list': [290.96388, 14.019167, 0.23425331666666666],
     'exclude': False,
     'name': 'circle',
     'params': [Number(290.96388), Number(14.019167), Ang(843.31194")]}
    """

    def __init__(self, shape_name, shape_params):
        self.name = shape_name
        self.params = shape_params

        self.comment = None
        self.exclude = False
        self.continued = False

    def __repr__(self):
        params_string = ",".join(map(repr, self.params))
        if self.exclude:
            return "Shape : -%s ( %s )" % (self.name, params_string)
        else:
            return "Shape : %s ( %s )" % (self.name, params_string)

    def set_exclude(self):
        self.exclude = True


class Property(object):
    def __init__(self, text):
        self.text = text

    def __repr__(self):
        return "Property : " + repr(self.text)


class CoordCommand(Property):
    def __repr__(self):
        return "CoordCommand : " + repr(self.text)


class Global(Property):
    def __repr__(self):
        return "Global : " + repr(self.text)


class Comment(Property):
    def __repr__(self):
        return "Comment : " + repr(self.text)


class RegionPusher(object):
    def __init__(self):
        self.flush()

    def flush(self):
        self.stack = []
        self.comment = None
        self.continued = None

    def pushAtom(self, s, l, tok):
        self.stack.append(tok[-1])

    def pushComment(self, s, l, tok):
        self.comment = tok[-1].strip()

    def set_continued(self, s, l, tok):
        self.continued = True



import copy

from pyparsing import (Literal, CaselessKeyword, Word, Optional, Combine,
                       ZeroOrMore, nums, alphas, And, Or, quotedString,
                       QuotedString, White)

from .region_numbers import CoordOdd, CoordEven, Distance, Angle

from .parser_helper import wcs_shape, define_shape_helper, Shape


def get_ds9_attr_parser():
    lhs = Word(alphas)
    paren = QuotedString("(", endQuoteChar=")")
    rhs = Or([Word(alphas + nums),
              Combine(Literal("#") + Word(alphas + nums)),  # color with '#'
              Combine(Word(alphas) + White() + Word(nums)),  # for point
              quotedString,
              QuotedString("{", endQuoteChar="}"),
              paren + ZeroOrMore(paren),
              Word(nums + " "),
              Word(nums + ".")
              ])
    expr = lhs + Optional(Literal("=").suppress() + rhs)
    expr.setParseAction(lambda s, l, tok: tuple(tok))

    return ZeroOrMore(expr)


ds9_shape_in_comment_defs = dict(
    text=wcs_shape(CoordOdd, CoordEven),
    vector=wcs_shape(CoordOdd, CoordEven,
                     Distance, Angle),
    composite=wcs_shape(CoordOdd, CoordEven, Angle),
    ruler=wcs_shape(CoordOdd, CoordEven, CoordOdd, CoordEven),
    compass=wcs_shape(CoordOdd, CoordEven, Distance),
    projection=wcs_shape(CoordOdd, CoordEven, CoordOdd, CoordEven, Distance),
    segment=wcs_shape(CoordOdd, CoordEven,
                      repeat=(0, 2))
)


class Ds9AttrParser(object):
    def set_continued(self, s, l, tok):
        self.continued = True

    def __init__(self):
        self.continued = False

        ds9_attr_parser = get_ds9_attr_parser()

        regionShape = define_shape_helper(ds9_shape_in_comment_defs)
        regionShape = regionShape.setParseAction(lambda s, l, tok: Shape(tok[0], tok[1:]))

        self.parser_default = ds9_attr_parser

        cont = CaselessKeyword("||").setParseAction(self.set_continued).suppress()
        line = Optional(And([regionShape, Optional(cont)])) + ds9_attr_parser

        self.parser_with_shape = line

    def parse_default(self, s):
        return self.parser_default.parseString(s)

    def parse_check_shape(self, s):
        l = self.parser_with_shape.parseString(s)
        if l and isinstance(l[0], Shape):
            if self.continued:
                l[0].continued = True
            return l[0], l[1:]
        else:
            return None, l


def get_attr(attr_list, global_attrs):
    """
    Parameters
    ----------
    attr_list : list
        A list of (keyword, value) tuple pairs
    global_attrs : tuple(list, dict)
        Global attributes which update the local attributes
    """
    local_attr = [], {}
    for kv in attr_list:
        keyword = kv[0]
        if len(kv) == 1:
            local_attr[0].append(keyword)
            continue
        elif len(kv) == 2:
            value = kv[1]
        elif len(kv) > 2:
            value = kv[1:]

        if keyword == 'tag':
            local_attr[1].setdefault(keyword, set()).add(value)
        else:
            local_attr[1][keyword] = value

    attr0 = copy.copy(global_attrs[0])
    attr1 = copy.copy(global_attrs[1])

    if local_attr[0]:
        attr0.extend(local_attr[0])

    if local_attr[1]:
        attr1.update(local_attr[1])

    return attr0, attr1



import numpy as np
from astropy.coordinates import SkyCoord
from astropy.wcs import WCS
from astropy.wcs.utils import proj_plane_pixel_scales


def _estimate_angle(angle, reg_coordinate_frame, header):
    """Transform an angle into a different frame

    Parameters
    ----------
    angle : float, int
        The number of degrees, measured from the Y axis in origin's frame

    reg_coordinate_frame : str
        Coordinate frame in which ``angle`` is defined

    header : `~astropy.io.fits.Header` instance
        Header describing the image

    Returns
    -------
    angle : float
        The angle, measured from the Y axis in the WCS defined by ``header'`
    """
    y_axis_rot = _calculate_rotation_angle(reg_coordinate_frame, header)
    return angle - y_axis_rot


def _calculate_rotation_angle(reg_coordinate_frame, header):
    """Calculates the rotation angle from the region to the header's frame

    This attempts to be compatible with the implementation used by SAOImage
    DS9. In particular, this measures the rotation of the north axis as
    measured at the center of the image, and therefore requires a
    `~astropy.io.fits.Header` object with defined 'NAXIS1' and 'NAXIS2'
    keywords.

    Parameters
    ----------
    reg_coordinate_frame : str
        Coordinate frame used by the region file

    header : `~astropy.io.fits.Header` instance
        Header describing the image

    Returns
    -------
    y_axis_rot : float
        Degrees by which the north axis in the region's frame is rotated when
        transformed to pixel coordinates
    """
    new_wcs = WCS(header)
    region_frame = SkyCoord(
        '0d 0d',
        frame=reg_coordinate_frame,
        obstime='J2000')
    region_frame = SkyCoord(
        '0d 0d',
        frame=reg_coordinate_frame,
        obstime='J2000',
        equinox=region_frame.equinox)

    origin = SkyCoord.from_pixel(
        header['NAXIS1'] / 2,
        header['NAXIS2'] / 2,
        wcs=new_wcs,
        origin=1).transform_to(region_frame)

    offset = proj_plane_pixel_scales(new_wcs)[1]

    origin_x, origin_y = origin.to_pixel(new_wcs, origin=1)
    origin_lon = origin.data.lon.degree
    origin_lat = origin.data.lat.degree

    offset_point = SkyCoord(
        origin_lon, origin_lat + offset, unit='degree',
        frame=origin.frame.name, obstime='J2000')
    offset_x, offset_y = offset_point.to_pixel(new_wcs, origin=1)

    north_rot = np.arctan2(
        offset_y - origin_y,
        offset_x - origin_x) / np.pi * 180.

    cdelt = new_wcs.wcs.get_cdelt()
    if (cdelt > 0).all() or (cdelt < 0).all():
        return north_rot - 90
    else:
        return -(north_rot - 90)



import copy
import warnings
from pyparsing import (Literal, CaselessKeyword, CaselessLiteral,
                       Optional, And, Or, Word,
                       StringEnd, ParseException, Combine,
                       alphas)
from .region_numbers import CoordOdd, CoordEven, Distance, Angle, Integer
from .parser_helper import (wcs_shape, define_shape_helper, Shape, Global,
                            RegionPusher, define_expr, define_line,
                            CoordCommand,
                            comment_shell_like, define_simple_literals)
from .ds9_attr_parser import Ds9AttrParser, get_attr
from .wcs_converter import (convert_to_imagecoord,
                            convert_physical_to_imagecoord)

ds9_shape_defs = dict(
    circle=wcs_shape(CoordOdd, CoordEven, Distance),
    rotbox=wcs_shape(CoordOdd, CoordEven, Distance, Distance, Angle,
                     repeat=(2, 4)),
    box=wcs_shape(CoordOdd, CoordEven, Distance, Distance, Angle,
                  repeat=(2, 4)),
    polygon=wcs_shape(CoordOdd, CoordEven, repeat=(0, 2)),
    ellipse=wcs_shape(CoordOdd, CoordEven, Distance, Distance, Angle, repeat=(2, 4)),
    annulus=wcs_shape(CoordOdd, CoordEven, Distance, repeat=(2, 3)),
    panda=wcs_shape(CoordOdd, CoordEven, Angle, Angle, Integer, Distance, Distance, Integer),
    pie=wcs_shape(CoordOdd, CoordEven,
                  Distance, Distance,
                  Angle, Angle),
    epanda=wcs_shape(CoordOdd, CoordEven,
                     Angle, Angle, Integer,
                     Distance, Distance, Distance,
                     Distance, Integer, Angle),
    bpanda=wcs_shape(CoordOdd, CoordEven,
                     Angle, Angle, Integer,
                     Distance, Distance, Distance,
                     Distance, Integer, Angle),
    point=wcs_shape(CoordOdd, CoordEven),
    line=wcs_shape(CoordOdd, CoordEven, CoordOdd, CoordEven),
    vector=wcs_shape(CoordOdd, CoordEven, Distance, Angle),
    text=wcs_shape(CoordOdd, CoordEven),
)

image_like_coordformats = ["image", "physical", "detector", "logical"]


class RegionParser(RegionPusher):

    def __init__(self):

        RegionPusher.__init__(self)

        self.shape_definition = ds9_shape_defs
        regionShape = define_shape_helper(self.shape_definition)
        regionShape = regionShape.setParseAction(lambda s, l, tok: Shape(tok[0], tok[1:]))

        regionExpr = define_expr(regionShape,
                                 negate_func=lambda s, l, tok: tok[-1].set_exclude(),
                                 )

        coord_command_keys = ['PHYSICAL', 'IMAGE', 'FK4', 'B1950', 'FK5',
                              'J2000', 'GALACTIC', 'ECLIPTIC', 'ICRS',
                              'LINEAR', 'AMPLIFIER', 'DETECTOR']

        coordCommandLiterals = define_simple_literals(coord_command_keys)
        coordCommandWCS = Combine(CaselessLiteral("WCS") + Optional(Word(alphas)))

        coordCommand = (coordCommandLiterals | coordCommandWCS)
        coordCommand.setParseAction(lambda s, l, tok: CoordCommand(tok[-1]))

        regionGlobal = comment_shell_like(CaselessKeyword("global"),
                                          lambda s, l, tok: Global(tok[-1]))

        regionAtom = (regionExpr | coordCommand | regionGlobal)

        regionAtom = regionAtom.setParseAction(self.pushAtom)

        regionComment = comment_shell_like(Literal("#"),
                                           parseAction=self.pushComment)

        line_simple = define_line(atom=regionAtom,
                                  separator=Literal(";"),
                                  comment=regionComment
                                  )

        line_w_composite = And([regionAtom,
                                CaselessKeyword("||").setParseAction(self.set_continued)
                                ]) \
                           + Optional(regionComment)

        line = Or([line_simple, line_w_composite])

        self.parser = Optional(line) + StringEnd()

    def parseLine(self, l):
        self.parser.parseString(l)
        s, c, continued = self.stack, self.comment, self.continued
        self.flush()

        return s, c, continued

    def parse(self, s):

        for l in s.split("\n"):
            try:
                s, c, continued = self.parseLine(l)
            except ParseException:
                warnings.warn("Failed to parse : " + l)
                self.flush()
                continue

            if len(s) > 1:
                for s1 in s[:-1]:
                    yield s1, None

                s[-1].comment = c
                s[-1].continued = continued
                yield s[-1], c
            elif len(s) == 1:
                s[-1].comment = c
                s[-1].continued = continued
                yield s[-1], c
            elif c:
                yield None, c

            self.flush()

    def convert_attr(self, l):
        global_attr = [], {}

        parser = Ds9AttrParser()

        for l1, c1 in l:
            if isinstance(l1, Global):
                for kv in parser.parse_default(l1.text):
                    if len(kv) == 1:
                        global_attr[0].append(kv[0])
                    elif len(kv) == 2:
                        if kv[0] == 'tag':
                            global_attr[1].setdefault(kv[0], set()).add(kv[1])
                        else:
                            global_attr[1][kv[0]] = kv[1]

            elif isinstance(l1, Shape):
                if c1:
                    attr_list = parser.parse_default(c1)
                    attr0, attr1 = get_attr(attr_list, global_attr)
                else:
                    attr0, attr1 = global_attr
                l1n = copy.copy(l1)
                l1n.attr = attr0, attr1
                yield l1n, c1

            elif not l1 and c1:
                shape, attr_list = parser.parse_check_shape(c1)
                if shape:
                    shape.attr = get_attr(attr_list, global_attr)
                    yield shape, c1
            else:
                yield l1, c1

    @staticmethod
    def sky_to_image(shape_list, header):
        """Converts a `ShapeList` into shapes with coordinates in image coordinates

        Parameters
        ----------
        shape_list : `pyregion.ShapeList`
            The ShapeList to convert
        header : `~astropy.io.fits.Header`
            Specifies what WCS transformations to use.

        Yields
        -------
        shape, comment : Shape, str
            Shape with image coordinates and the associated comment

        Note
        ----
        The comments in the original `ShapeList` are unaltered

        """

        for shape, comment in shape_list:
            if isinstance(shape, Shape) and \
                    (shape.coord_format not in image_like_coordformats):

                new_coords = convert_to_imagecoord(shape, header)

                l1n = copy.copy(shape)

                l1n.coord_list = new_coords
                l1n.coord_format = "image"
                yield l1n, comment

            elif isinstance(shape, Shape) and shape.coord_format == "physical":

                if header is None:
                    raise RuntimeError("Physical coordinate is not known.")

                new_coordlist = convert_physical_to_imagecoord(shape, header)

                l1n = copy.copy(shape)

                l1n.coord_list = new_coordlist
                l1n.coord_format = "image"
                yield l1n, comment

            else:
                yield shape, comment

    def filter_shape(self, sss):
        return [s1[0] for s1 in sss if isinstance(s1[0], Shape)]

    @staticmethod
    def filter_shape2(sss):
        r = [s1 for s1 in sss if isinstance(s1[0], Shape)]
        return zip(*r)



import numpy as np
import pyregion._region_filter as region_filter
import warnings


def as_region_filter(shape_list, origin=1):
    """
    Often, the regions files implicitly assume the lower-left corner
    of the image as a coordinate (1,1). However, the python convetion
    is that the array index starts from 0. By default (origin = 1),
    coordinates of the returned mpl artists have coordinate shifted by
    (1, 1). If you do not want this shift, use origin=0.
    """

    filter_list = []
    for shape in shape_list:

        if shape.name == "composite":
            continue

        if shape.name == "polygon":
            xy = np.array(shape.coord_list) - origin
            f = region_filter.Polygon(xy[::2], xy[1::2])

        elif shape.name == "rotbox" or shape.name == "box":
            xc, yc, w, h, rot = shape.coord_list
            # -1 for change origin to 0,0
            xc, yc = xc - origin, yc - origin

            f = region_filter.Rotated(region_filter.Box(xc, yc, w, h),
                                      rot, xc, yc)

        elif shape.name == "ellipse":
            xc, yc = shape.coord_list[:2]
            # -1 for change origin to 0,0
            xc, yc = xc - origin, yc - origin
            angle = shape.coord_list[-1]

            maj_list, min_list = shape.coord_list[2:-1:2], shape.coord_list[3:-1:2]

            if len(maj_list) > 1:
                w1, h1 = max(maj_list), max(min_list)
                w2, h2 = min(maj_list), min(min_list)

                f1 = region_filter.Ellipse(xc, yc, w1, h1) \
                     & ~region_filter.Ellipse(xc, yc, w2, h2)
                f = region_filter.Rotated(f1, angle, xc, yc)
            else:
                w, h = maj_list[0], min_list[0]
                f = region_filter.Rotated(region_filter.Ellipse(xc, yc, w, h),
                                          angle, xc, yc)

        elif shape.name == "annulus":
            xc, yc = shape.coord_list[:2]
            # -1 for change origin to 0,0
            xc, yc = xc - origin, yc - origin
            r_list = shape.coord_list[2:]

            r1 = max(r_list)
            r2 = min(r_list)

            f = region_filter.Circle(xc, yc, r1) & ~region_filter.Circle(xc, yc, r2)

        elif shape.name == "circle":
            xc, yc, r = shape.coord_list
            # -1 for change origin to 0,0
            xc, yc = xc - origin, yc - origin

            f = region_filter.Circle(xc, yc, r)

        elif shape.name == "panda":
            xc, yc, a1, a2, an, r1, r2, rn = shape.coord_list
            # -1 for change origin to 0,0
            xc, yc = xc - origin, yc - origin

            f1 = region_filter.Circle(xc, yc, r2) & ~region_filter.Circle(xc, yc, r1)
            f = f1 & region_filter.AngleRange(xc, yc, a1, a2)

        elif shape.name == "pie":
            xc, yc, r1, r2, a1, a2 = shape.coord_list
            # -1 for change origin to 0,0
            xc, yc = xc - origin, yc - origin

            f1 = region_filter.Circle(xc, yc, r2) & ~region_filter.Circle(xc, yc, r1)
            f = f1 & region_filter.AngleRange(xc, yc, a1, a2)

        elif shape.name == "epanda":
            xc, yc, a1, a2, an, r11, r12, r21, r22, rn, angle = shape.coord_list
            # -1 for change origin to 0,0
            xc, yc = xc - origin, yc - origin

            f1 = region_filter.Ellipse(xc, yc, r21, r22) & ~region_filter.Ellipse(xc, yc, r11, r12)
            f2 = f1 & region_filter.AngleRange(xc, yc, a1, a2)
            f = region_filter.Rotated(f2, angle, xc, yc)
            # f = f2 & region_filter.AngleRange(xc, yc, a1, a2)

        elif shape.name == "bpanda":
            xc, yc, a1, a2, an, r11, r12, r21, r22, rn, angle = shape.coord_list
            # -1 for change origin to 0,0
            xc, yc = xc - origin, yc - origin

            f1 = region_filter.Box(xc, yc, r21, r22) & ~region_filter.Box(xc, yc, r11, r12)
            f2 = f1 & region_filter.AngleRange(xc, yc, a1, a2)
            f = region_filter.Rotated(f2, angle, xc, yc)
            # f = f2 & region_filter.AngleRange(xc, yc, a1, a2)

        else:
            warnings.warn("'as_region_filter' does not know how to convert {0}"
                          " to a region filter.".format(shape.name))
            continue

        if shape.exclude:
            filter_list = [region_filter.RegionOrList(*filter_list) & ~f]
        else:
            filter_list.append(f)

    return region_filter.RegionOrList(*filter_list)
